;;;;
;;;; Transaction control needed to speed things up:
;;;;  http://www.sqlite.org/faq.html#q19
;;;;
;;;; select distinct card.answer from card inner join tag_card on (card.id = tag_card.card_id) inner join tag on (tag_card.tag_id = tag.id) where tag.tag = 10;
;;;;
;;;; select * from card as w1 inner join tag_card as kw1 on (w1.id = kw1.card_id) inner join tag as k1 on (kw1.tag_id = k1.id) where k1.tag >= 10 and w1.id not in (select w2.id from card as w2 inner join tag_card as kw2 on (w2.id = kw2.card_id) inner join tag as k2 on (kw2.tag_id = k2.id) where k2.tag <= 9);
;;;;

;; (quicklisp:quickload 'sqlite)
;; (quicklisp:quickload 'cl-fad)
;; (quicklisp:quickload 'cl-ppcre)
;; (quicklisp:quickload 'alexandria)
;; (quicklisp:quickload 'day-to-day)

(require 'sqlite)
(require 'cl-fad)
(require 'cl-ppcre)
(require 'alexandria)
;;(require 'cl-json)
(require 'day-to-day) ; TODO
(defpackage :wrenshoe-wrangle
  (:use :cl
	:cl-fad
	:day-to-day
	:sqlite))

(in-package :wrenshoe-wrangle)

;;;
;;; Handy utilities for converting old files and data into the
;;; environment.
;;;

(defun parse-cards-text-file (file)
  "Get the formatted tag text that we'll need from the
file. Tab-separated withthree or four columns. First is kana question,
second is (tag) writing, third is commentary, (optional) fourth is
example."
  (mapcar #'(lambda (x)
	      (list :question (first x)
		    :answer (second x)
		    :commentary (third x)
		    :examples (fourth x))) ; will be split to "example" table
	  (filter
	   (file->lines file :function
		       #'(lambda (x) (cl-ppcre::split "\\t+" x)))   
	   #'(lambda (x) (cl-ppcre::scan "^\\s*\#" (car x))) ; not a comment
	   #'(lambda (x) (< (length x) 2))))) ; at least 2 columns (tag, imi)

(defun parse-levels-text-file (levels-file)
  "Parse trivial format, single characters separated by nothing, new
lines to separate lists. Each level is a list."
  (let ((ldata '())
	(level-data (file->lines levels-file :function
				 #'(lambda (x) (cl-ppcre::split "" x)))))
    (loop 
       for level-section in level-data
       for level from 10 downto 0 ; easiest level is 10, so work from there
       do (dolist (question level-section)
	    ;;(setf ldata (cons (cons question level) ldata))))
	    (setf ldata (cons (list :question question :level level) ldata))))
    ldata))

;;;
;;; Database (sqlite3) helper functions. Probably will spin them out
;;; into their own class sometime...
;;;

(defun ->create-schema (db)
  "Create the schema."
  ;; info
  (execute-non-query db "create table if not exists info (id integer primary key, name text not null, description text not null, question_label text not null, answer_label text not null, commentary_label text not null)")
  ;; card
  (execute-non-query db "create table if not exists card (id integer primary key, question text not null, answer text not null, commentary text null)")
  ;; example
  (execute-non-query db "create table if not exists example (id integer primary key, card_id integer not null, text text not null)")
  ;; tag_set
  (execute-non-query db "create table if not exists tag_set (id integer primary key, name text not null, description text null, type text not null)")
  ;; tag
  (execute-non-query db "create table if not exists tag (id integer primary key, name text not null, description text null, value text not null, tag_set_id integer not null)")
  ;; tag_card
  (execute-non-query db "create table if not exists tag_card (id integer primary key, tag_id integer, card_id integer)"))

(defun ->info (db name desc qlabel alabel clabel)
  (execute-non-query
   db "insert into info (name, description, question_label, answer_label, commentary_label) values (?, ?, ?, ?, ?)"
   name desc qlabel alabel clabel))

(defun ->card (db question answer commentary)
  (execute-non-query
   db "insert into card (question, answer, commentary) values (?, ?, ?)"
   question answer commentary))

(defun ->example (db card-id text)
  (execute-non-query
   db "insert into example (card_id, text) values (?, ?)"
   card-id text))

(defun ->tag-set (db tag name desc)
  (execute-non-query
   db "insert into tag_set (tag, name, description) values (?, ?, ?)"
   tag name desc))

(defun ->tag (db answer tag)
  (execute-non-query
   db "insert into tag (question, tag) values (?, ?)"
   answer tag))

(defun ->tag-card (db tag-id card-id)
  (execute-non-query
   db "insert into tag_card (tag_id, card_id) values (?, ?)"
   tag-id card-id))

(defun <-last-insert-row-id (db)
  "The the id of the last inserted row."
  (last-insert-rowid db))

(defun <-select-template (db table &rest fields)
  "...";
  (mapcar #'(lambda (row)
	      (loop
		 for f in fields
		 for i in row
		 collect (cons (intern (string-upcase f) "KEYWORD") i)))
	  (execute-to-list db
	   (d2d:ccat "select " (d2d:join-strings fields :with ", ")
		     " from " table))))

(defun <-select-all-info (db)
  "Create a plist of all info data."
  (<-select-template db "info" "id" "name" "description" "question_label" "answer_label" "commentary_label"))

(defun <-select-all-cards (db)
  "Create a plist of all cards."
  (<-select-template db "card" "id" "question" "answer" "commentary"))

(defun <-select-all-examples (db)
  "Create a plist of all examples."
  (<-select-template db "example" "id" "card_id" "text"))

(defun <-select-all-tag-sets (db)
  "Create a plist of all tag_sets."
  (<-select-template db "tag_set" "id" "name" "description" "type"))

(defun <-select-all-tags (db)
  "Create a plist of all tag_sets."
  (<-select-template db "tag" "id" "name" "description" "value" "tag_set_id"))

(defun <-select-all-tag-cards (db)
  "Create a plist of all tag_sets."
  (<-select-template db "tag_card" "id" "tag_id" "card_id"))

;;;
;;; Resources are a collection of things like data paths, etc., that
;;; make compiling the exportable data easy.
;;;
;;; Define a resource. These classes will try and simplify as much as
;;; possible, but in the end you will have to provide your own
;;; mappings and functions to make it all work out.
;;;

;; Define resource names.
(defvar *file-ext* ".sexpr")
(defvar *db-file-ext* ".db")
(defvar *info-file* (d2d:ccat "info" *file-ext*))
(defvar *cards-file* (d2d:ccat "cards" *file-ext*))
(defvar *garbage-file* (d2d:ccat "garbage" *file-ext*))
(defvar *tag-set-pre* "tag_set.")

(defclass resource ()
  ((name
    :accessor name
    :initform nil
    :initarg :name)
   (path
    :accessor path
    :initform nil
    :initarg :path)
   (built?
    :accessor built?
    :initform nil
    :initarg :built)
   (tags-hash
    :accessor tags-hash
    :initform nil)
   (garbage-hash
    :accessor garbage-hash
    :initform nil)))

(defclass resource-trivial (resource)
  ())

(defclass resource-complex (resource)
  ())

(defun pull-tag-sets (location)
  "Get a listing of the tag set files in the d2d current directory."
  (let ((regexp (d2d:ccat *tag-set-pre* "(.*)" *file-ext* "$")))
    (d2d:filter (d2d:ls location) (lambda (x)
			   (not (cl-ppcre:scan-to-strings
				 regexp (namestring x)))))))

(defun identify-resource (name)
  "Either return a symbol that identifies a named resource, or nil."
  (let ((info-location (merge-pathnames (here name) *info-file*))
	(cards-location (merge-pathnames (here name) *cards-file*))
	(tag-set-locations (pull-tag-sets (here name))))
    (cond
      ((and (file-exists-p info-location)
	    (file-exists-p cards-location)
	    tag-set-locations)
       'resource-complex)
      ((and (file-exists-p info-location)
	    (file-exists-p cards-location))
       'resource-trivial)
      (t nil))))

;; Example:
;;  (cd "wrenshoe/data")
;;  (assemble (make-resource "jamo"))
(defun make-resource (name)
    "Resource class constructor. This will handle for trivial and
complex resources, auto-detecting which one to try."
    (cond
      ((eq 'resource-trivial (identify-resource name))
       (make-instance 'resource-trivial :name name :path (pwd)))
      ((eq 'resource-complex (identify-resource name))
       (make-instance 'resource-complex :name name :path (pwd)))
      (t nil)))

;;;
;;; Methods shared for all resource types.
;;;

;; BUG: how to fix tag files?
(defmethod data-path ((r resource) &optional (symb nil))
  "Returns base data path by default; locations of other things too..."
  (let ((base (merge-pathnames (pathname-as-directory (name r))
			       (pathname-as-directory (path r)))))
    (cond
      ((eq symb 'garbage)
       (merge-pathnames base (pathname-as-file *garbage-file*)))
      ((eq symb 'db)
       (merge-pathnames base (pathname-as-file (ccat (name r) *db-file-ext*))))
      ((eq symb 'info)
       (merge-pathnames base (pathname-as-file *info-file*)))
      ((eq symb 'cards)
       (merge-pathnames base (pathname-as-file *cards-file*)))
      ;; ((eq symb 'tags)
      ;;  (merge-pathnames base (pathname-as-file *tags-file*)))
      (t base))))
  
;; ;; ...
;; (defmethod boom-data ((r resource))
;;   "Reset the database to nothing. This includes destroying it if it exists."
;;   (let ((path (data-path r 'db)))
;;     (when (file-exists-p path)
;;       (delete-file path)))
;;   (setf (slot-value r 'tags-hash) nil)
;;   (setf (slot-value r 'garbage-hash) nil))

;; Should be good for all types.
(defmethod reset-data ((r resource))
  "Add the schema to the resource db, creating it in the process. t or nil."
  (let ((path (data-path r 'db)))
    (when (file-exists-p path)
      (delete-file path)))
  (setf (slot-value r 'built?) nil)
  (setf (slot-value r 'tags-hash) nil)
  (setf (slot-value r 'garbage-hash) nil)
  ;;(boom-data r)
  (let ((path (data-path r 'db)))
    (with-open-database (db path)
      (with-transaction db
	(->create-schema db))))
  t)

;; Should be good for all types.
(defmethod load-info ((r resource))
  "Load the info data into the (extant) database--info and tag_set tables."
  (let* ((info (marshal (data-path r 'info)))
	 (name (or (getf info :NAME) ""))
	 (desc (or (getf info :DESCRIPTION) ""))
	 (qlabel (or (getf info :QUESTION_LABEL) "Question"))
	 (alabel (or (getf info :ANSWER_LABEL) "Answer"))
	 (clabel (or (getf info :COMMENTARY_LABEL) "Commentary")))
    ;; (tag-set-list (or (getf info :TAGS) nil)))
    (with-open-database (db (data-path r 'db))
      (with-transaction db
	;; Add to info.
	(->info db name desc qlabel alabel clabel))))
  t)
;; 	;; Add to tag_set.
;; 	(dolist (entry tag-set-list)
;; 	  (->tag-set db
;; 		     (getf entry :tag)
;; 		     (getf entry :name) ; possible nil implied
;; 		     (getf entry :description))))) ; possible nil implied
;; t))

;; Should be good for all types.
;; TODO:
(defmethod export-to-json ((r resource))
  "Export all of the contents of the database to a fun JSON format. Only after assemble though."
  (when (built? r)
    (with-open-database (db (data-path r 'db))
	(list 
	 (cons :info (car (<-select-all-info db)))
	 (cons :cards (<-select-all-cards db))
	 (cons :examples (<-select-all-examples db))
	 (cons :tag-cards (<-select-all-tag-cards db))
	 (cons :tags (<-select-all-tags db))
	 (cons :tag-sets (<-select-all-tag-sets db))))))

;;;
;;; Specialized methods for resource-trivial.
;;;

;; Combines: load-cards load-tags and bridge into one function
;; FUNction!
(defmethod load-cards ((r resource-trivial))
  "Load the cards data into the (extant) database."
  (with-open-database (db (data-path r 'db))
    (with-transaction db
      (dolist (card (marshal (data-path r 'cards)))
	(let ((question (getf card :question))
	      (answer (getf card :answer))
	      (commentary (getf card :commentary))
	      ;;(tag (or (getf card :tag) 0)) ; 0 as default
	      (card-id nil))
	      ;;(tag-id nil))
	  ;; Do main card. Capture last insert id.
	  (->card db question answer commentary)
	  (setf card-id (<-last-insert-row-id db))
	  ;; Add examples, if any.
	  (let ((ex (getf card :examples)))
	    (when ex
	      (let ((lid (<-last-insert-row-id db)))
		(dolist (e ex)
		  (->example db lid e)))))
	  ;; ;; Add tag. Capture last insert id.
	  ;; (->tag db answer tag)
	  ;; (setf tag-id (<-last-insert-row-id db))
	  ;; ;; Bridge tag and card.
	  ;; (->tag-card db tag-id card-id)
	))))t)

(defmethod assemble ((r resource-trivial))
  "Smart enough to do the right things depending on the presence or
absence of files."  (and
   (ll "Reset. ") (reset-data r) ; ...
   (ll "Info. ") (load-info r)
   (ll "Cards. ") (load-cards r))
  (setf (slot-value r 'built?) t))

;;;
;;; Special methods for resource-complex.
;;;

(defmethod load-cards ((r resource-complex))
  "Load the cards data into the (extant) database."
  (with-open-database (db (data-path r 'db))
    (with-transaction db
      (dolist (card (marshal (data-path r 'cards)))
	(->card db (getf card :question) (getf card :answer)
		(getf card :commentary))
	(let ((examples (getf card :examples)))
	  (when examples
	    (let ((last-id (<-last-insert-row-id db)))
	      (dolist (examples examples)
		(->example db last-id examples))))))))
  t)

(defmethod load-tags ((r resource-complex))
  "Load the cards data into the (extant) database. Also builds the tags hash."
  (let ((db-location (data-path r 'db))
	(tags-location (data-path r 'tags)))
    ;; Make sure that the db and tags file both exist.
    (when (and (file-exists-p db-location) (file-exists-p tags-location))
      ;; Ready hash.
      (setf (slot-value r 'tags-hash) (make-hash-table :test 'equal))
      (let ((tags-hash (slot-value r 'tags-hash)))
	;; Ready db.
	(with-open-database (db db-location)
	  (with-transaction db
	    ;; Walk through and add to db and hash.
	    (dolist (e (marshal (data-path r 'tags)))
	      (let ((in-question (getf e :question))
		    (in-tag (getf e :tag)))
		(setf (gethash in-question tags-hash) (<-last-insert-row-id db))
		(->tag db in-question in-tag)))))
	tags-hash))))


;; Read in hiragana/katakana/garbage and cache 'em.
(defmethod load-garbage ((r resource-complex))
  ;; Ready hash.
  (let ((db-location (data-path r 'db))
	(garbage-location (data-path r 'garbage))
	(garbage-hash (slot-value r 'garbage-hash)))
    (when (and (file-exists-p db-location) (file-exists-p garbage-location))
      (setf garbage-hash (make-hash-table :test 'equal))
      (dolist (piece (marshal (data-path r 'garbage)))
	(setf (gethash piece garbage-hash) t)))
    garbage-hash))


(defmethod bridge-tags-cards ((r resource-complex))
  ;; Read in tags and garbage.
  (let ((tags-hash (slot-value r 'tags-hash))
  	(garbage-hash (slot-value r 'garbage-hash)))
    (with-open-database (db (data-path r 'db))
      (with-transaction db
	;; Read in cards, explode, check hash, then make tags-card link.
	(dolist (card (<-select-all-cards db))
	  (let ((card-id (getf card :ID)) (answer (getf card :ANSWER)))
	    (dolist (question (cl-ppcre::split "" answer))
	      ;; Make sure it is not a uninteresting character.
	      (when (or (not garbage-hash)
			(not (gethash question garbage-hash)))
		;; ...
		(let ((tags-id (gethash question tags-hash)))
		  (cond
		    (tags-id ; if the tag was found
		     ;; Insert in bridge table.
		     (->tag-card db tags-id card-id))
		    (t ; if it's not a found tag, add it as tag 0
		     (progn
		       ;; Add tag to db.
		       (->tag db question 0)
		       ;; Get inserted id (shadow eariler tag id).
		       (let ((tags-id (<-last-insert-row-id db)))
			 ;; Add to cache.
			 (setf (gethash question tags-hash) tags-id)
			 ;; Insert in bridge table.
			 (->tag-card db tags-id card-id))))))))))))))

(defmethod assemble ((r resource-complex))
  "Smart enough to do the right things depending on the presence or
absence of files."
  (and
   (ll "Reset. ") (reset-data r) ; ...
   (ll "Info. ") (load-info r)
   (ll "Cards. ") (load-cards r) ; load cards--most basic and necessary
   (and
    (ll "Tags. ") (load-tags r) ; 
    (progn (ll "Garbage? ") (load-garbage r) t) ; wrapped since it's optional
    (ll "Bridge. ") (bridge-tags-cards r))))
