####
####
####

package Wrenshoe::Compile;

use strict;
use utf8;
use Data::Dumper;
use Try::Tiny;
use File::Slurp;
use Text::CSV;
use JSON::PP; # JSON::XS was dumb and made utf-8 neigh impossible
use Wrenshoe::Utilities qw(:all);
use Wrenshoe::Deck;
use Wrenshoe::Error;

## Every parse call will reset the deck with a new one.
sub new {

  my $class = shift;
  my $self = {};

  ## Our card deck slot.
  #my $self->{deck} = undef;
  $self->{deck} = Wrenshoe::Deck->new();

  bless $self, $class;
  return $self;

}

## TSV filename.
sub parse_filename {
  my $self = shift;
  my $filename = shift || die "need a filename arg";

  ## Technically unecessary right now since parse_lines will reset as
  ## well...
  $self->{deck} = Wrenshoe::Deck->new();

  if( ! -f $filename ){
    WrenshoeDataError->throw(resource=>$filename, type=>'file',
			     message=>'file not found');
  }

  my $lines = [];
  # try {
  #$lines = read_file($filename, array_ref => 1, chomp => 1);
  $lines = slurp_to_lines($filename);
  # }catch{
  #   WrenshoeDataError->throw(resource=>$filename, type=>'file',
  # 			     message=>'file not usable')
  # };

  $self->parse_lines($lines);
}

## CSV filename.
sub parse_csv_filename {
  my $self = shift;
  my $filename = shift || die "need a filename arg";

  ## Technically unecessary right now since parse_lines will reset as
  ## well...
  $self->{deck} = Wrenshoe::Deck->new();

  if( ! -f $filename ){
    WrenshoeDataError->throw(resource=>$filename, type=>'file',
			     message=>'file not found');
  }

  my $lines = [];
  my $csv = Text::CSV->new({ sep_char => ',' });
  open(FILE, '<', $filename) or die "Could not open '$filename' $!\n";
  while (my $line = <FILE>){
    chomp $line;
    if( $csv->parse($line) ){
      my @csv_fields = $csv->fields();
      push @$lines, join("\t", @csv_fields);
    }else{
      push @$lines, '';
    }
  }
  close(FILE);

  $self->parse_lines($lines);
}

## JSON set filenames.
sub parse_json_fileset {
  my $self = shift;
  my $meta_filename = shift || die "need a meta filename arg";
  my $data_filename = shift || die "need a data filename arg";

  ## Technically unecessary right now since parse_lines will reset as
  ## well...
  $self->{deck} = Wrenshoe::Deck->new();

  ## Double check the files.
  if( ! -f $meta_filename ){
    WrenshoeDataError->throw(resource=>$meta_filename, type=>'file',
			     message=>'file not found');
  }
  if( ! -f $data_filename ){
    WrenshoeDataError->throw(resource=>$data_filename, type=>'file',
			     message=>'file not found');
  }

  ## Read in meta file.
  open(META, $meta_filename) || die "Can't open meta file: $!";
  my $meta_json = '';
  while( <META> ){
    if( $_ =~ /^\#/ ){
      ## Ignore if comment. Yes, technically not JSON.
    }else{
      $meta_json .= $_;
    }
  }
  close META;

  ## JSON into a meta object.
  my $js = JSON::PP->new();
  my $meta_data = {};
  try{
    $meta_data = $js->decode($meta_json);
  }catch{
    #  ll('mjson: ' . $meta_json);
    #  ll('mdata: ' . Dumper($meta_data));
    #   #ll('mj: ' . $_);
    #   #ll('mj: ' . @_);
    #   #ll('mj: ' . $@);
    WrenshoeDataError->throw(resource=>$meta_filename, type=>'file',
			     message=>'file not correct JSON');
  };

  ## Cheap: convert the meta-object into lines.
  my $lines = [];
  push @$lines, $meta_data->{name};
  push @$lines, $meta_data->{description};
  my $types = [];
  my $names = [];
  foreach my $col (@{$meta_data->{columns}}){
    push @$types, $col->{type};
    push @$names, $col->{name};
  }
  push @$lines, join("\t", @$types);
  push @$lines, join("\t", @$names);

  #ll("lines (meta):\n" . join("\n", @$lines));

  ## Now load the data file as lines.
  open(DATA, $data_filename) || die "Can't open data file: $!";
  while( <DATA> ){
    chomp($_);
      push @$lines, $_;
  }
  close DATA;

  #ll("lines (final):\n" . join("\n", @$lines));

  $self->parse_lines($lines, 4);
}

## This is the core function--most handlers will try and get something
## that can run through here--it takes an aref of strings like:
## 
sub parse_lines {
  my $self = shift;
  my $strs_list = shift || die "need a lines arg";
  ## Optional correction for when the metadata is appended for
  ## parsing.
  my $line_shift = shift || 0;

  $self->{deck} = Wrenshoe::Deck->new();

  my $line_number = 0;
  my $meta_line_number = 1;
  my $number_of_cards = 0;
  my $stored_type_line = undef; # row 3 for later processing
  my $stored_name_line = undef; # row 4 for later processing

  #ll('parse ' . scalar(@$strs_list) . ' line(s)');
  #ll('here (t): ' . ref($strs_list));
  foreach my $line (@$strs_list){

    ## Advance file/warning index.
    #$line_number += 1;
    $line_number++;

    #ll('@' . $line_number ': ' . $line );

    ## Remove surrounding /spaces/--tabs are still informational and
    ## important, but we can afford to be a little forgiving.
    chomp($line);
    $line = strip($line, "\ ");

    ## Skip line if it looks like it could be a comment.
    if( $line =~ /^#/ ){
      next;
    }

    ## Again, being a little flexible--skip lines that have nothing in
    ## them at all.
    if( $line eq '' ){
      next;
    }

    #ll('passed drops: ' . $line_number);

    ## We need to process the few real lines we see as the header so
    ## we have meta-information to work with for the rest of the file.
    if( $meta_line_number == 1 ){ # quiz title line
      $meta_line_number++;
      $self->{deck}->name($line);
      next;
    }elsif( $meta_line_number == 2 ){ # quiz description line
      $meta_line_number++;
      $self->{deck}->description($line);
      next;
    }elsif( $meta_line_number == 3 ){ # get field type line
      $meta_line_number++;
      $stored_type_line = $line;
      next;
    }elsif( $meta_line_number == 4 ){ # last meta line; get field name line
      $stored_name_line = $line;
      $meta_line_number++;

      ## Okay, we have the types and names; let's make sure they match
      ## and them try and add them as faces.
      my @raw_types = splode($stored_type_line);
      my @raw_names = splode($stored_name_line);

      if( scalar(@raw_types) > scalar(@raw_names) ){
	## Bad.
	#ll('more types than names: ' . $line_number);
	WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				    type=>'header',
				    message=>'more types than names');
      }elsif( scalar(@raw_types) < scalar(@raw_names) ){
	## Bad.
	#ll('more names than types: ' . $line_number);
	WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				    type=>'header',
				    message=>'more names than types');
      }else{

	#ll('matched names and types: ' . $line_number);
	#ll('types count: ' . scalar(@raw_types));
	#ll('names count: ' . scalar(@raw_names));
	#ll("types:\n" . Dumper(\@raw_types));
	#ll("names:\n" . Dumper(\@raw_names));

	## Since they are the same length, process them at the same
	## time and make sure they are taken correctly by add_face.
	my $findex = 0;
	foreach my $raw_type (@raw_types){
	  my $raw_name = $raw_names[$findex];

	  my $nname = strip($raw_name);
	  my $ntype = strip($raw_type);

	  $self->{deck}->add_face($nname, $ntype);
	  if( ! $self->{deck}->okay() ){
	    ## Error.
	    my $ferror_type = $self->{deck}->error_type();
	    my $ferror_msg = $self->{deck}->error_message();
	    ll('BAD!: ' . $line_number);
	    WrenshoeCompileError->throw(line=>($line_number - $line_shift),
					type=>$ferror_type,
					message=>$ferror_msg);
	  }

	  $findex++;
	}
      }

      next;
    }

    #ll('passed meta: ' . $line_number);

    ## Okay, we have a line of raw data, so let's break it apart,
    ## check it against the column counts, and place it into cards.
    my @raw_data = splode($line);
    #ll('sploded card: ' . scalar(@raw_data) . ' (' . $line . ')');
    if( scalar(@raw_data) == 0 ){
      ## Error.
      #ll('nothing for: ' . $line);
      WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				  type=>'data',
				  message=>'no usable data');
    }

    ## Now try and add this as a card, and check the results to see if
    ## there was an error.
    $number_of_cards = $self->{deck}->add_card(\@raw_data);
    if( ! $self->{deck}->okay() ){
      ## Error.
      ##ll('unable to add: ' . $line . ' @' . $line_number);
      #die 'unable to add: ' . $line;
      my $ferror_type = $self->{deck}->error_type();
      my $ferror_msg = $self->{deck}->error_message();
      WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				  type=>$ferror_type,
				  message=>$ferror_msg);
    }

    #ll('added: ' . $line_number);
  }
  #ll('out: ' . $line_number);

  ## Now let's do a retrospective on what we should have seen.
  if( $line_number == 0 ){
    WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				type=>'file',
				message=>'input seems to be empty');
  }
  if( $meta_line_number < 5 ){
    WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				type=>'header',
				message=>'not enough header information');
  }
  if( $number_of_cards == 0 ){
    WrenshoeCompileError->throw(line=>($line_number - $line_shift),
				type=>'data',
				message=>'no card data');
  }
}

##
sub deck {
  my $self = shift;
  return $self->{deck};
}

##
sub deck_object {
  my $self = shift;
  return $self->{deck}->object();
}


1;
