####
#### The fundamental unit in Wrenshoe handling.
####
#### Data holder with operations to get stuff in and out.
####
#### So we don't have to do things like rethrow a millions kinds of
#### errors upstream/elsewhere, we'll follow an agent-style for error
#### handling.
####
#### Error types are 'tag', 'column', and 'entry'
####

package Wrenshoe::Deck;

use strict;
use utf8;
use Data::Dumper;
use Data::UUID;
use Clone qw(clone);
#use JSON::PP;

use Wrenshoe::Utilities qw(:all);

sub new {

  ##
  my $class = shift;
  my $self = {};

  ## A UUID.
  my $uuids = Data::UUID->new();
  my $uuid = $uuids->to_string( $uuids->create() );

  $self->{id} = $uuid;
  $self->{name} = 'A Wrenshoe Quiz';
  $self->{description} = 'A simple quiz.';
  #$self->{column_types} = []; # a temp data structure while parsing
  $self->{columns} = [];
  $self->{tag_fields} = {}; # tag name to count map
  $self->{tags} = {}; # literal tag to count map

  ## Our main data.
  $self->{cards} = [];

  ## Make sure we are fixed once we start adding cards.
  $self->{card_lock} = 0;

  ## Error handling like an agent.
  $self->{okay} = 1;
  $self->{error_type} = undef;
  $self->{error_message} = undef;

  bless $self, $class;
  return $self;
}

sub _error {
  my $self = shift;
  my $type = shift || "need type arg";
  my $msg = shift || "need message arg";

  $self->{okay} = 0;
  $self->{error_type} = $type;
  $self->{error_message} = $msg;
}

sub okay {
  my $self = shift;
  return $self->{okay};
}

sub error_type {
  my $self = shift;
  return $self->{error_type};
}

sub error_message {
  my $self = shift;
  return $self->{error_message};
}

sub reset_error {
  my $self = shift;
  $self->{okay} = 1;
  $self->{error_type} = undef;
  $self->{error_message} = undef;
}

sub id {
  my $self = shift;
  my $arg = shift || undef;
  return getter_setter($self, 'id', $arg);
}

sub name {
  my $self = shift;
  my $arg = shift || undef;
  return getter_setter($self, 'name', $arg);
}

sub description {
  my $self = shift;
  my $arg = shift || undef;
  return getter_setter($self, 'description', $arg);
}

sub get_cards(){
  my $self = shift;
  return clone($self->{cards});
}
sub get_columns(){
  my $self = shift;
  return clone($self->{columns});
}
sub get_tags(){
  my $self = shift;
  return clone($self->{tags});
}
sub get_tag_fields(){
  my $self = shift;
  return clone($self->{tag_fields});
}

## Helper.
sub _increment_tag_field {
  my $self = shift;
  my $tag_field = shift || die "need arg";

  if( ! defined( $self->{tag_fields}{$tag_field}) ){
    $self->{tag_fields}{$tag_field} = 0;
  }
  $self->{tag_fields}{$tag_field}++;

  return $self->{tag_fields}{$tag_field};
}

##
sub object {
  my $self = shift;

  my $out_obj =
    {
     'id' => $self->{id},
     'name' => $self->{name},
     'description' => $self->{description},
     'columns' => $self->{columns},
     'cards' => $self->{cards},
     'tag_fields' => $self->{tag_fields},
     'tags' => $self->{tags}
    };

  return clone($out_obj);
}

# ##
# sub json {
#   my $self = shift;
#   my $pretty_p = shift || 0;

#   my $retstr = undef;
#   my $json = JSON::PP->new();
#   if( $pretty_p ){
#     $json->pretty();
#   }
#   $retstr = $json->encode($self->object());

#   return $retstr;
# }

## Helper.
sub _increment_tag {
  my $self = shift;
  my $tag = shift || die "need arg";

  if( ! defined( $self->{tags}{$tag}) ){
    $self->{tags}{$tag} = 0;
  }
  $self->{tags}{$tag}++;

  return $self->{tags}{$tag};
}

sub add_face {
  my $self = shift;
  my $name = shift || undef;
  my $type = shift || 'face';

  if( ! defined($name) || $name eq '' ){
    $self->_error('column', 'has no name');
    $name = '???';
  }

  ## Cannot add faces once we have started adding cards.
  if( $self->{card_lock} ){
    $self->_error('column', 'attempt to add face after adding card');
  }elsif( $type ne 'face' && $type ne 'tag' ){
    ## Currently, type must be 'face' or 'tag'.
    $self->_error('column',  $type . ' is not a known type');
  }

  my $face =
    {
     'name' => $name,
     'type' => $type
    };

  push @{$self->{columns}}, $face;

  return scalar(@{$self->{columns}});
}

## Actually does a fair amount of lifting.
## Two possible exceptions:
## 
sub add_card {
  my $self = shift;
  my $card = shift || die "need arg";

  ## An error if cards and columns do not match.
  my $card_num = scalar(@$card);
  my $col_num = scalar(@{$self->{columns}});
  if( $card_num > $col_num ){
    $self->_error('entry', 'card has more faces then there are columns ('.
		  $card_num .' to '. $col_num .')');
  }else{

    ## Now we know that we have $card_num <= $col_num. In the case
    ## that $card_num < $col_num, we want to load those extra cards
    ## with blanks just to be nice--some readers and parsers have
    ## funny ideas (looking at you Google Spreadsheet API).
    if( $card_num < $col_num ){
      for( my $s = ($col_num - $card_num); $s > 0; $s-- ){
    	$col_num++;
    	push @$card, '';
      }
    }

    ## Cycle through and add things according to the data structure
    ## layout. Since both are the same length, we'll cycle through the
    ## columns.
    my $card_struct = {};
    my $index = 0;
    foreach my $col_struct (@{$self->{columns}}){

      ## The column bits.
      my $cname = $col_struct->{name};
      my $ctype = $col_struct->{type};

      ## The card bits. Can often be nothing.
      my $card_entry = strip($$card[$index] || '');

      ## There are only tags and faces.
      if( $ctype eq 'face' ){
	## Easy if we are just dealing with a face.
	$card_struct->{$cname} = $card_entry;
      }else{ #if( $ctype eq 'tag' ){

	my $tag_entry = $card_entry;

	###
	### Tag entries, OTOH, are a little messy since they are required
	### to be defined and can be comma-delimited.
	###

	## A tag column must meaningfully exist. Fail out if
	## necessary.
	if( ! $tag_entry ){
	  $self->_error('tag', 'missing tag entry (maybe empty)');
	  return scalar(@{$self->{cards}});
	}else{

	  my @raw_tags = splode($tag_entry, ',');
	  foreach my $raw_tag (@raw_tags){
	    my $tag = strip($raw_tag);

	    ## Each tag must meaningfully exist.
	    if( ! $tag ){
	      $self->_error('tag', 'missing tag (maybe empty)');
	    }else{

	      ## Make sure that our tagspace is defined.
	      ## Then add our tag to it.
	      if( ! defined($card_struct->{_tags}) ){
		$card_struct->{_tags} = {};
	      }
	      $card_struct->{_tags}{$tag} = $cname;

	      ## Now add the meta-data for the tag.
	      $self->_increment_tag_field($cname);
	      $self->_increment_tag($tag);
	    }
	  }
	}
      }

      $index++;
    }
    ## For safe keeping.
    push @{$self->{cards}}, $card_struct;
  }


  ## Now that we have successfully added a card, make sure no columns
  ## are messed with.
  if( ! $self->{card_lock} ){
    $self->{card_lock} = 1;
  }

  return scalar(@{$self->{cards}});
}



1;
