####
#### Super set and emitter for decks.
####

package Wrenshoe::DeckSet;

use strict;
use utf8;
use Data::Dumper;
use JSON::PP;
use LWP::UserAgent;
#use XML::Simple qw(:strict);
use XML::TreePP;

use Wrenshoe::Utilities qw(:all);
use Wrenshoe::Error;
use Wrenshoe::Deck;
use Wrenshoe::Compile;

sub new {

  ##
  my $class = shift;
  my $self = {};

  $self->{decks} = [];

  bless $self, $class;
  return $self;
}

sub count {
  my $self = shift;
  return scalar(@{$self->{decks}});
}

sub add_deck {
  my $self = shift;
  my $deck = shift || "need deck arg";
  push @{$self->{decks}}, $deck->object();
  return scalar(@{$self->{decks}});
}

## Looking for a "Published to the web" URL.
## Must be "Published on the web" to work, see:
## https://developers.google.com/google-apps/spreadsheets/#spreadsheets_api_urls_visibilities_and_projections
## API: https://developers.google.com/google-apps/documents-list/#DownloadingSpreadsheets
## https://code.google.com/p/gdata-issues/issues/detail?id=1816
## https://code.google.com/a/google.com/p/apps-api-issues/issues/detail?id=1813
sub add_google_spreadsheet {
  my $self = shift;
  my $in_url = shift || die "need a url arg";

  ## Prep our agent--we'll hopefully use it a bunch.
  my $ua = LWP::UserAgent->new();
  $ua->timeout(5);

  # ## Technically unecessary right now since parse_lines will reset as
  # ## well...
  # $self->{deck} = Wrenshoe::Deck->new();

  ## First, do what we can to extract the key.
  #my $url = 'https://docs.google.com/spreadsheet/pub?key='0AoMWJeL67IDbdGNCdUlkQWo5ZlRKRk5CVTNzMm9CYXc'&output=txt';

  ## Get key--starts at key, ends at '&' (not \w).
  $in_url =~ /key=(\w+)/;
  my $key = $1;
  if( ! $key || $key eq '' ){
    WrenshoeDataError->throw(resource=>$in_url, type=>'URL',
			     message=>'could not extract key from URL');
  }
  #ll('key: ' . $key);

  ## Get meta-info.
  my $get_url = 'https://spreadsheets.google.com/feeds/worksheets/'.
    $key. '/public/basic';
  ## Direct TSV?
  # my $get_url = 'https://docs.google.com/spreadsheet/pub?key=' .
  #   $key . '&output=txt&sheet=od7';
 #my $get_url = 'http://spreadsheets.google.com/feeds/download/spreadsheets/Export?key=' . $key. '&hl=en&exportFormat=text&gid=0';
  ## Use an agent to try and grab the sheet.
  my $doc = undef;
  my $response = $ua->get($get_url);
  if( ! $response->is_success() ){
    WrenshoeDataError->throw(resource=>$get_url, type=>'URL',
			     message=> 'could not retrieve Google spreadsheet '
			     . 'at "' . $get_url. '" '
			     . '(' . $response->status_line() . ')');
  }

  ## XML doc is somewhat okay--extract the entry items of possible.
  $doc = $response->decoded_content();
  my $tpp = XML::TreePP->new();
  my $mhash = $tpp->parse($doc);
  #die "mhash:\n" . Dumper($mhash);
  # my $xs = XML::Simple->new(KeyAttr => { server => 'name' },
  # 			    ForceArray => [ 'server', 'address']);
  # my $mhash = $xs->XMLin($doc);
  #ll("doc:\n" . Dumper($mhash));
  #ll("here");
  if( ! $mhash || ! $mhash->{feed} || ! $mhash->{feed}{entry} ){
    WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			     'could not read Google spreadsheet meta data');
  }
  #ll("there");

  my $entries = [];
  my $entry_type = ref($mhash->{feed}{entry});
  if( $entry_type eq 'ARRAY' ){
    ## Is list.
    $entries = $mhash->{feed}{entry};
  }elsif( $entry_type eq 'HASH' ){
    ## Is a singleton.
    $entries = [$mhash->{feed}{entry}];
  }else{
    WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			     'unknown type for entries');
  }

  ## Cycle through the entry or entries and try to extract the
  ## worksheet ids. If we find worksheet ids (proxy ids), convert them
  ## into real gids as best we can using the table. Google is an
  ## idiot.
  #ll("zher");
  my $found_links = [];
  foreach my $entry (@$entries){
    #ll('entry: ' . $entry);

    if( ! $entry->{id} ){
      WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			       'could not find Google spreadsheet entry id');
    }

    ## Okay, now recover the gid proxy.
    $entry->{id} =~ /\/public\/basic\/(.*)$/;
    my $gid_proxy = $1;
    if( ! $gid_proxy ){
      WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			       'could not extract plausible worksheet id');
    }

    ## Assemble the final download URL.
    my $gid_table =
      {
       'od6'=>0, 'od7'=>1, 'od4'=>2, 'od5'=>3, 'oda'=>4,
       'odb'=>5, 'od8'=>6, 'od9'=>7, 'ocy'=>8, 'ocz'=>9,
       'ocw'=>10, 'ocx'=>11, 'od2'=>12, 'od3'=>13, 'od0'=>14,
       'od1'=>15, 'ocq'=>16, 'ocr'=>17, 'oco'=>18, 'ocp'=>19,
       'ocu'=>20, 'ocv'=>21, 'ocs'=>22, 'oct'=>23, 'oci'=>24,
       'ocj'=>25, 'ocg'=>26, 'och'=>27, 'ocm'=>28, 'ocn'=>29,
       'ock'=>30, 'ocl'=>31, 'oe2'=>32, 'oe3'=>33, 'oe0'=>34,
       'oe1'=>35, 'oe6'=>36, 'oe7'=>37, 'oe4'=>38, 'oe5'=>39,
       'odu'=>40, 'odv'=>41, 'ods'=>42, 'odt'=>43, 'ody'=>44,
       'odz'=>45, 'odw'=>46, 'odx'=>47, 'odm'=>48, 'odn'=>49,
       'odk'=>50, 'odl'=>51, 'odq'=>52, 'odr'=>53, 'odo'=>54,
       'odp'=>55, 'ode'=>56, 'odf'=>57, 'odc'=>58, 'odd'=>59,
       'odi'=>60, 'odj'=>61, 'odg'=>62, 'odh'=>63, 'obe'=>64,
       'obf'=>65, 'obc'=>66, 'obd'=>67, 'obi'=>68, 'obj'=>69,
       'obg'=>70, 'obh'=>71, 'ob6'=>72, 'ob7'=>73, 'ob4'=>74,
       'ob5'=>75, 'oba'=>76, 'obb'=>77, 'ob8'=>78, 'ob9'=>79,
       'oay'=>80, 'oaz'=>81, 'oaw'=>82, 'oax'=>83, 'ob2'=>84,
       'ob3'=>85, 'ob0'=>86, 'ob1'=>87, 'oaq'=>88, 'oar'=>89,
       'oao'=>90, 'oap'=>91, 'oau'=>92, 'oav'=>93, 'oas'=>94,
       'oat'=>95, 'oca'=>96, 'ocb'=>97, 'oc8'=>98, 'oc9'=>99
      };
    my $gid = $gid_table->{$gid_proxy};
    if( ! defined($gid) ){
      WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			       'could not map worksheet id to gid');
    }

    ## Okay, final URL.
    my $furl =
      # 'https://docs.google.com/feeds/download/spreadsheets/Export?key=' .
      # 	$key . '&exportFormat=tsv&gid='. $gid;
      'https://docs.google.com/spreadsheet/pub?key=' . $key .
	'&output=txt&pub=1&gid=' . $gid;
    #ll($furl);
    push @$found_links, $furl;
  }
  #ll("where");

  ## Make sure that we got something.
  if( scalar(@$entries) == 0 ){
    WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			     'unable to read any spreadsheet');
  }

  ## Make sure that we have the same number of recovered links as
  ## entries.
  if( scalar(@$found_links) < scalar(@$entries) ){
    WrenshoeDataError->throw(resource=>$get_url, type=>'URL', message=>
			     'unable to read all spreadsheet worksheets');
  }

  ## Okay, now we have two nice URLs, lets, resolve them.
  #ll('top');
  foreach my $tsv_url (@$found_links){

    ## Try
    my $response = $ua->get($tsv_url);
    if( ! $response->is_success() ){
      WrenshoeDataError->throw(resource=>$tsv_url, type=>'URL',
			       message=> 'could not retrieve Google worksheet' .
			       ' (' . $response->status_line() . ')');
    }

    ## Got a TSV doc!
    my $tsv_doc = $response->decoded_content();
    #ll("tsv_url: " . $tsv_url);
    my @tsv_lines = split("\n", $tsv_doc);
    #ll("tsv_doc: " . scalar(@tsv_lines));
    my $c = Wrenshoe::Compile->new();

    $c->parse_lines(\@tsv_lines);

    ## Remeber--we push objects, not the actual decks.
    push @{$self->{decks}}, $c->deck()->object();
  }

  return scalar(@{$self->{decks}});
}

##
sub object {
  my $self = shift;
  return $self->{decks};
}

##
sub json {
  my $self = shift;
  my $pretty_p = shift || 0;

  my $retstr = undef;
  my $json = JSON::PP->new();
  if( $pretty_p ){
    $json->pretty(1);
  }
  $retstr = $json->encode($self->object());

  return $retstr;
}

##
sub data_js {
  my $self = shift;
  my $pretty_p = shift || 0;

  my $data = $self->json($pretty_p);
  my @out = ();
  push @out, 'var wrenshoe = {};';
  push @out, 'wrenshoe.current_state = null;';
  push @out, 'wrenshoe.current_deck = null;';
  push @out, 'wrenshoe.current_front_faces = [];';
  push @out, 'wrenshoe.current_back_faces = [];';
  push @out, 'wrenshoe.current_filters = [];';
  push @out, 'wrenshoe.current_meta_show_p = false;';
  push @out, 'wrenshoe.deck_list = ' . $data . ';';
  return join("\n", @out);
}



1;
