use strict;
use Test::More 'no_plan';
use Data::Dumper;
use Try::Tiny;

use Wrenshoe::Compile;
use Wrenshoe::Error;
use Wrenshoe::Utilities qw(:all);

## Universal setup.
my $c = Wrenshoe::Compile->new();
my $deck = undef; # will reuse this one a lot
my $triggered_p = 0; # will reuse this one a lot
my $test_name = '???';

## Test empty state.
{
  $test_name = '0';
  $triggered_p = 0;

  #$deck = $c->deck()->get_columns();
  is(scalar(@{$c->deck()->get_columns()}), 0, "$test_name no cols yet");
  is(scalar(@{$c->deck()->get_cards()}), 0, "$test_name no cards yet");
}

## Test file not there.
{
  $test_name = '00';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/this_file_so_does_not_exist');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeDataError') ){
      is($_->type, 'file', "$test_name non-extant file");
    }else{
      fail("$test_name bad error on no file");
    }
  }finally{
    fail("$test_name skipped error on no file") if ! $triggered_p;
  };
}

## Test 01a--check that direct access works as expected.
{
  $test_name = '01a';
  $triggered_p = 0;

  open(FILE, '<', 'data/test01.wrenshoe') or die "Cannot open $test_name: $!";
  my @strs01a = <FILE>;
  close FILE;

  try{
    $c->parse_lines(\@strs01a);
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 0, "$test_name lines: 0");
      is($_->type, 'file', "$test_name empty file");
    }else{
      #err('foo1: ' . $c->deck()->okay());
      #err('foo2: ' . scalar(@strs01a));
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 01b--now same thing using the file name.
{
  $test_name = '01b';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test01.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 0, "$test_name lines: 0");
      is($_->type, 'file', "$test_name empty file");
    }else{
      #err('foo1: ' . $_);
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 02.
{
  $test_name = '02';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test02.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 4, "$test_name lines: 4");
      is($_->type, 'header', "$test_name no header");
      is($_->message, 'not enough header information', "$test_name no header");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 03.
{
  $test_name = '03';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test03.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 8, "$test_name lines: 8");
      is($_->type, 'data', "$test_name no data");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };

  is(scalar(@{$c->deck()->get_columns()}), 4, "$test_name still got 4 cols");
  my $cols = $c->deck()->get_columns();
  ok($$cols[0]->{name} eq 'title1', "$test_name title1");
  ok($$cols[0]->{type} eq 'face', "$test_name face 1");
  ok($$cols[1]->{name} eq 'title2', "$test_name title2");
  ok($$cols[1]->{type} eq 'face', "$test_name face 2");
  ok($$cols[2]->{name} eq 'title3', "$test_name title3");
  ok($$cols[2]->{type} eq 'face', "$test_name face 3");
  ok($$cols[3]->{name} eq 'title4', "$test_name title4");
  ok($$cols[3]->{type} eq 'tag', "$test_name face 4");
}

## Test 04.
{
  $test_name = '04';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test04.wrenshoe');
  }catch{
    $triggered_p = 1;
    #err('foo1: ' . $c->deck()->okay());
    #err('foo2: ' . $_->message);
    #err('foo3: ' . scalar(@{$c->deck()->get_cards()}));
  }finally{
    fail("unexpected error $test_name") if $triggered_p;
  };

  is(scalar(@{$c->deck()->get_columns()}), 4, "$test_name got 4 cols");
  is(scalar(@{$c->deck()->get_cards()}), 3, "$test_name got 4 cards");
  ok($c->deck()->name() eq 'uber title', "$test_name got title");
  ok($c->deck()->description() eq 'desc', "$test_name got description");
  my $cards = $c->deck()->get_cards();
  ok($$cards[2]->{'title3'} eq 'fooC3', "$test_name card title check");
}

## Test 05.
{
  $test_name = '05';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test05.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 9, "$test_name bad line 9");
      is($_->type, 'tag', "$test_name bad tag");
      is($_->message, 'missing tag entry (maybe empty)', "$test_name message");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 06.
{
  $test_name = '06';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test06.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 10, "$test_name bad line 10");
      is($_->type, 'tag', "$test_name bad tag");
      is($_->message, 'missing tag (maybe empty)', "$test_name message");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 07.
{
  $test_name = '07';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test07.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 11, "$test_name bad line 11");
      is($_->type, 'tag', "$test_name bad tag");
      is($_->message, 'missing tag (maybe empty)', "$test_name message");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 08.
{
  $test_name = '08';
  $triggered_p = 0;

  try{
    $c->parse_filename('data/test08.wrenshoe');
  }catch{
    $triggered_p = 1;
    if( $_->isa('WrenshoeCompileError') ){
      is($_->line, 10, "$test_name bad entry (missing) 10");
      is($_->type, 'tag', "$test_name bad tag, not entry, due to right-fill");
      is($_->message, 'missing tag entry (maybe empty)', "$test_name message");
    }else{
      fail("bad error $test_name");
    }
  }finally{
    fail("skipped error $test_name") if ! $triggered_p;
  };
}

## Test 09--semi-real using morse meta-JSON and data files.
{
  $test_name = '09';
  $triggered_p = 0;

  try{
    $c->parse_json_fileset('data/test09.wmeta', 'data/test09.wdata');
  }catch{
    $triggered_p = 1;
    #err('deck->ok: ' . $c->deck()->okay());
    #err('foo2: ' . $_->message);
  }finally{
    fail("unexpected error $test_name") if $triggered_p;
  };

  is(scalar(@{$c->deck()->get_columns()}), 2, "$test_name got 2 cols");
  is(scalar(@{$c->deck()->get_cards()}), 3, "$test_name got 3 cards");
  ok($c->deck()->name() eq 'Morse Code', "$test_name got title");
  ok($c->deck()->description() eq 'ITU Morse code.',
     "$test_name got description");
  # my $obj = $c->deck()->object();
  # is_deeply($obj,
  # 	    {
  # 	    }
  # # ok($$cards[2]->{'title3'} eq 'fooC3', "$test_name card title check");
}

## Test 10--CSV.
{
  $test_name = '10';
  $triggered_p = 0;

  try{
    $c->parse_csv_filename('data/test10.lo-csv.wrenshoe');
  }catch{
    $triggered_p = 1;
  }finally{
    fail("unexpected error $test_name") if $triggered_p;
  };

  is(scalar(@{$c->deck()->get_columns()}), 3, "$test_name got 4 cols");
  is(scalar(@{$c->deck()->get_cards()}), 2, "$test_name got 4 cards");
  my $cards = $c->deck()->get_cards();
  ok($$cards[0]->{'title3'} eq '', "$test_name no card title");
  ok($$cards[1]->{'title3'} eq 'bing', "$test_name card title");
}
