#!/usr/bin/python
# -*- coding: utf-8 -*-
####
#### Parsing a different format than the perl CLI directory-level
#### compiler (compile-directory.pl). The 
####
#### # Legit input:
#### title
#### description
#### face     face        tag
#### title1   title2      title3
#### fooA1    fooA2       fooA3
#### fooB1                fooB3
####          fooC2       fooC3, barC3
####

import sys
import io
import codecs
import re
import uuid
import json
import csv

# class WrenshoeFileError(Exception):
#     def __init__(self, reason, in_type='unknown'):
#         self.reason = reason
#         self.type = in_type
#     def __str__(self):
#         return repr(self.reason + '(type: ' + self.type + ')')

class WrenshoeCompileError(Exception):
    def __init__(self, reason, line):
        self.reason = reason
        self.line = line
    def __str__(self):
        return repr(self.reason + '(line: ' + self.line + ')')

class WrenshoeCompiler:
    
    ## ...
    def __init__(self):

        ## Our main data.
        self.cards = []

        ## Other things that we'll need.
        self.id = str(uuid.uuid4())
        self.name = 'A Wrenshoe Quiz'
        self.description = 'This quiz was generated by the Wrenshoe file compiler.'
        self.column_types = [] # a temp data structure while parsing
        self.columns = []
        self.tag_fields = {} # tag name to count map
        self.tags = {} # literal tag to count map

    def increment_tag_field(self, tag_field):
        if not tag_field in self.tag_fields:
            self.tag_fields[tag_field] = 0
        else:
            self.tag_fields[tag_field] += 1

    def increment_tag(self, tag):
        if not tag in self.tags:
            self.tags[tag] = 0
        else:
            self.tags[tag] += 1

    ## 
    def parse_filename(self, fname):
        with codecs.open(fname, 'r', encoding='utf_8_sig') as f:
            self.parse_file(f)

    ## Take a file as a string filename.
    ## Does some autodetection and conversion.
    def parse_file(self, file):

        csv_string_list = []
        tsv_string_list = []

        ## First try as CSV.        
        reader = csv.reader(file, delimiter=',')
        for row in reader:
            #print("row: " + row)
            csv_string_list.append("\t".join(row))

        ## Reset for another try.
        file.seek(0)

        ## Try again with just tab.
        for line in file:
            tsv_string_list.append(line)

        ## 
        tsv_tab_count = 0;
        tsv_comma_count = 0;
        for tsv_string in tsv_string_list:
            tsv_tab_count += tsv_string.count("\t")
            tsv_comma_count += tsv_string.count(",")
        csv_tab_count = 0;
        csv_comma_count = 0;
        for csv_string in csv_string_list:
            csv_tab_count += csv_string.count("\t")
            csv_comma_count += csv_string.count(",")

        ## Run the one that looks better.
        # print("")
        # print("tsv_comma_count: " + str(tsv_comma_count))
        # print("tsv_tab_count: " + str(tsv_tab_count))
        # print("csv_comma_count: " + str(csv_comma_count))
        # print("csv_tab_count: " + str(csv_tab_count))
        if tsv_tab_count >= csv_comma_count:
            self.parse_tsv_list(tsv_string_list)
        else:
            #print("run: tsv")
            self.parse_tsv_list(csv_string_list)

    ## Take a file as a string of tab separated values
    def parse_tsv_list(self, strs):

        ## Reusable regexps.
        comment_pattern = re.compile('^\#+')
        tab_split_re = re.compile('\t')
        #pipe_split_re = re.compile('\|\|')
        #comma_split_re = re.compile(',')
        tag_entry_split_re = re.compile(',')

        ## For all of the lines in the file (which we are assuming is
        ## encoded in utf-8).
        line_number = 0;
        meta_line_number = 1;

        for line in strs:
            
            ## Advance file/warning index.
            line_number += 1

            ## Remove surrounding /spaces/--tabs are still
            ## informational and important, but we can afford to
            ## be a little forgiving.
            line = line.rstrip('\n')
            line = line.strip(' ') 

            ## Skip if it looks like it could be a comment.
            if comment_pattern.search(line):
                continue

            ## Again, being a little flexible...
            if line == '':
                continue

            ## We need to process the first real line we see
            ## as the header so we have meta-information to
            ## work with for the rest of the file.
            if meta_line_number == 1: # quiz title line
                meta_line_number += 1
                self.name = line
                continue
            elif meta_line_number == 2: # quiz description line
                meta_line_number += 1
                self.description = line
                continue
            elif meta_line_number == 3: # field type line
                meta_line_number += 1
                self.process_type_line(line, line_number)
                continue
            elif meta_line_number == 4: # field name line
                meta_line_number += 1
                self.process_name_line(line, line_number)
                continue

            ## Okay, we have a line of raw data, so let's break it
            ## apart, check it against the column counts, and
            ## place it into cards.
            raw_data = tab_split_re.split(line)
            if len(raw_data) == 0:
                raise WrenshoeCompileError('no usable data', line_number)

            ## Now let's check to make sure it conforms to our
            ## gross expectations from the header data.
            if len(raw_data) < len(self.columns):
                raise WrenshoeCompileError('row does not have enough entries to match header information', line_number)
            if len(raw_data) > len(self.columns):
                raise WrenshoeCompileError('row has more entries than header information', line_number)

            ## Looking good, now let's examine it a bit closer.
            column_index = 0
            card = {}
            card_tags = {}
            for raw_entry in raw_data:                
                entry = raw_entry.strip() # finally, remove cruft

                ## Recover our reference.
                col_name = self.columns[column_index]['name']
                col_type = self.columns[column_index]['type']

                if col_type == 'face':
                    ## Faces can be anything really, even empty.
                    card[col_name] = entry
                elif col_type == 'tag':
                    ## Tags must always be defined. If defined,
                    ## comma separated.
                    if entry == '':
                        raise WrenshoeCompileError('tag entry must always be defined', line_number)
                        
                    ## Split on comma and check.
                    tag_data = tag_entry_split_re.split(entry)
                    if len(tag_data) == 0:
                        raise WrenshoeCompileError('tag entries must always be defined', line_number)

                    ## Examine each individual tag.
                    for raw_entry_tag in tag_data:
                        entry_tag = raw_entry_tag.strip()

                        if len(entry_tag) == '':
                            raise WrenshoeCompileError('tag entry in tag column cannot be empty', line_number)
                        
                        ## Finally, a kernel of goodness. Also
                        ## increment the tag meta-data.
                        card_tags[entry_tag] = col_name
                        self.increment_tag_field(col_name)
                        self.increment_tag(entry_tag)

                else:
                    raise WrenshoeCompileError('impossible', line_number)

                ## On to the next column...
                column_index += 1

            ## Assemble into self.cards.
            card['_tags'] = card_tags
            self.cards.append(card)

        ## If we got out of our input with nothing, that is a problem.
        if line_number == 0:
            raise WrenshoeCompileError('input seems empty', line_number)
        if line_number != 0 and len(self.cards) == 0:
            raise WrenshoeCompileError('no data found', line_number)
        if line_number == 1:
            raise WrenshoeCompileError('no data, only header', line_number)

    ## This will properly set: column_types. If
    ## anything goes wrong, it will throw an exception.
    def process_type_line(self, line, line_number=1):

        tab_split_re = re.compile('\t')

        ## First, get rid of anything we had before.
        self.column_types = []

        ## Break down the header on tabs.
        line = line.strip(' ')
        raw_face_data = tab_split_re.split(line)
        if len(raw_face_data) == 0:
            raise WrenshoeCompileError('no usable face data', line_number)

        ## It looks like we got something, so lets process the header
        ## data.
        for raw_face_entry in raw_face_data:
            face_entry = raw_face_entry.strip(' ')
            
            ## We /must/ have a sensible type.
            if face_entry != 'face' and face_entry != 'tag':
                raise WrenshoeCompileError('type line has bad entry',
                                           line_number)

            ## We have a well-defined face and type--store them for
            ## later.
            self.column_types.append(face_entry)

        if len(self.column_types) == 0 :
            raise WrenshoeCompileError('type line appears to be empty',
                                       line_number)

    ## This will properly set: columns. If anything goes wrong, it
    ## will throw an exception.
    def process_name_line(self, line, line_number=1):

        tab_split_re = re.compile('\t')

        ## First, get rid of anything we had before.
        self.columns = []
        self.tag_fields = {} # ?
        self.tags = {} # ?

        ## Break down the name line on tabs.
        line = line.strip(' ')
        raw_name_data = tab_split_re.split(line)
        if len(raw_name_data) == 0:
            raise WrenshoeCompileError('no usable name line data', line_number)

        ## It looks like we got something, so lets process the header
        ## data.
        names = []
        for raw_name_entry in raw_name_data:
            name_entry = raw_name_entry.strip(' ')
            names.append(name_entry)

            if len(names) > len(self.column_types):
                  raise WrenshoeCompileError('too many names in name line',
                                             line_number)

            ## We have a well-defined face and type--store them for
            ## later.
            self.columns.append({'name': name_entry,
                                 'type': self.column_types[len(names) -1]})

        if len(self.columns) == 0 :
            raise WrenshoeCompileError('name line appears to be empty',
                                       line_number)

    def to_json(self, pretty_p=False):
        out_obj = {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'columns': self.columns,
            'cards': self.cards
            }
        if pretty_p:
            return json.dumps(out_obj, indent=4, sort_keys=True)
        else:
            return json.dumps(out_obj, sort_keys=True)

###
### CLI unit testing.
### Simulate with: from wrenshoecompiler import WrenshoeCompiler
###
if __name__ == '__main__':

    import unittest

    class TestEmpty(unittest.TestCase):
        
        def test(self):

            c = WrenshoeCompiler()
            self.assertEqual(len(c.cards), 0)
            self.assertEqual(len(c.columns), 0)
            self.assertEqual(len(c.to_json()), 174)
    
    class Test01a(unittest.TestCase):
        
        def test(self):

            c = WrenshoeCompiler()
            f = codecs.open('../../perl/lib/t/data/test01.wrenshoe', 'r', encoding='utf_8_sig')
            strs = f.readlines()
            try:
                c.parse_tsv_list(strs)
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'input seems empty')
                self.assertEqual(e.line, 0)
                
    class Test01b(unittest.TestCase):
        
        def test(self):

            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test01.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'input seems empty')
                self.assertEqual(e.line, 0)
    
    class Test02(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test02.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'no data found')
                self.assertEqual(e.line, 4)
    
    class Test03(unittest.TestCase):
        
        def test(self):

            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test03.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'no data found')
                self.assertEqual(e.line, 8)

            self.assertEqual(len(c.columns), 4)
            self.assertEqual(c.columns[0]['name'], 'title1')
            self.assertEqual(c.columns[0]['type'], 'face')
            self.assertEqual(c.columns[1]['name'], 'title2')
            self.assertEqual(c.columns[1]['type'], 'face')
            self.assertEqual(c.columns[2]['name'], 'title3')
            self.assertEqual(c.columns[2]['type'], 'face')
            self.assertEqual(c.columns[3]['name'], 'title4')
            self.assertEqual(c.columns[3]['type'], 'tag')
    
    class Test04(unittest.TestCase):
        
        def test(self):

            c = WrenshoeCompiler()
            c.parse_filename('../../perl/lib/t/data/test04.wrenshoe')

            self.assertEqual(len(c.columns), 4)
            self.assertEqual(len(c.cards), 3)
            self.assertEqual(c.name, 'uber title')
            self.assertEqual(c.description, 'desc')
            self.assertEqual(c.cards[2]['title3'], 'fooC3')
                
    class Test05(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test05.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'tag entry must always be defined')
                self.assertEqual(e.line, 9)

    class Test06(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test06.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'tag entry must always be defined')
                self.assertEqual(e.line, 8)

    class Test07(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            try:
                c.parse_filename('../../perl/lib/t/data/test07.wrenshoe')
            except WrenshoeCompileError as e:
                self.assertEqual(e.reason, 'tag entry must always be defined')
                self.assertEqual(e.line, 8)

    class Test10(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            c.parse_filename('../../perl/lib/t/data/test10.lo-csv.wrenshoe')
            self.assertEqual(len(c.columns), 3)
            self.assertEqual(len(c.cards), 2)
            self.assertEqual(c.cards[0]['title3'], '')
            self.assertEqual(c.cards[1]['title3'], 'bing')

    class Test11(unittest.TestCase):
        
        def test(self):
            c = WrenshoeCompiler()
            c.parse_filename('../../perl/lib/t/data/test11.lo-csv.wrenshoe')
            self.assertEqual(len(c.columns), 3)
            self.assertEqual(len(c.cards), 2)
            self.assertEqual(c.cards[0]['title3'], '')
            self.assertEqual(c.cards[1]['title3'], 'bing')

    unittest.main()
