#!/usr/bin/env perl
use Mojolicious::Lite;

use Try::Tiny;
use Template;
use Cwd qw(abs_path);

use Wrenshoe::Error;
use Wrenshoe::Compile;
use Wrenshoe::DeckSet;
use Wrenshoe::Page;

use GD::Barcode::QRcode;


# # Documentation browser under "/perldoc"
# plugin 'PODRenderer';

###
### Helper functions.
###

sub _insufficient_args_redirect {
  my $app = shift;

  my $title = shift || undef;
  my $description = shift || undef;
  my $url = shift || undef;

  ## Redirect on insufficient arguments.
  my $ret = 0;
  if( ! $title || ! $description || ! $url ){
    my $redo_url = $app->url_for("/input");
    $app->redirect_to($redo_url->query('title' => $title,
				       'description' => $description,
				       'url' => $url));
    $ret = 1;
  }
  return $ret;
}

sub _error {
  my $app = shift;

  my $type = shift || 'unknown';
  my $title = shift || 'Error';
  my $description = shift || 'Something bad happened :(';

  # my $error_url = $app->url_for("/error");
  # $app->redirect_to($error_url->query('type' => $type,
  # 				      'title' => $title,
  # 				      'description' => $description));
  $app->stash('TYPE' => $type);
  $app->stash('TITLE' => $title);
  $app->stash('DESCRIPTION' => $description);

  $app->render('error');
  return 1;
}

sub _assemble_attempt {
  my $ds = shift || die;
  my $url = shift || die;

  my $error_message = undef;
  try{
    $ds->add_google_spreadsheet($url);
  }catch{
    ## TODO: Use nice templates for this type of more complcated
    ## error.
    if( $_->isa('WrenshoeDataError') ){
      $error_message = "There was a " . $_->type . " error in your input.";
    }elsif( $_->isa('WrenshoeCompileError') ){
      $error_message = "There was a " . $_->type . " error in your input.";
    }elsif( $_->isa('WrenshoeError') ){
      $error_message = "There was a generic error in your input.";
    }else{
      $error_message = "Unknown error: $!";
    }
  };

  return $error_message;
}

sub _bookmark {
  my $app = shift;

  my $type = shift || die;
  my $title = shift || die;
  my $description = shift || die;
  my $url = shift || die;

  my $quiz_url = $app->url_for('/' . $type);
  my $quiz_url_str = $quiz_url->query('title' => $title,
				      'description' => $description,
				      'url' => $url);
  return $quiz_url_str;
}

###
### Routing.
###

## Landing and welcome page.
get '/' => sub {
  my $self = shift;
  $self->render('landing');
};

## Detailed information and contact.
get '/about' => sub {
  my $self = shift;
  $self->render('about');
};

## Detailed information and contact.
get '/learn' => sub {
  my $self = shift;
  $self->render('learn');
};

get '/input' => sub {
  my $self = shift;

  my $title = $self->param('title') || undef;
  my $description = $self->param('description') || undef;
  my $url = $self->param('url') || undef;
  my $submitted_p = $self->param('submitted_p') || undef;

  $self->stash('TITLE' => $title);
  $self->stash('DESCRIPTION' => $description);
  $self->stash('URL' => $url);
  $self->stash('SUBMITTED_P' => $submitted_p);

  ## Check title.
  my $bad_title_message = undef;
  if( ! defined($title) || $title eq '' ){
    $bad_title_message = 'You need to add your own title.';
  }
  $self->stash('BAD_TITLE_MESSAGE' => $bad_title_message);

  ## Check description.
  my $bad_description_message = undef;
  if( ! defined($description) || $description eq '' ){
    $bad_description_message = 'You need to add your own description.';
  }
  $self->stash('BAD_DESCRIPTION_MESSAGE' => $bad_description_message);

  ## Check URL.
  my $bad_url_message = undef;
  if( ! defined($url) || $url eq '' ){
    $bad_url_message = 'You need to add the URL to your Google Spreadsheet.';
  }
  $self->stash('BAD_URL_MESSAGE' => $bad_url_message);

  # ## Make note if this is a first run and block error messages if so.
  # my $first_run_p = 1;
  # if( defined($title) || defined($desc) || defined($url) ){
  #    $first_run_p = 0;
  # }
  # $self->stash('FIRST_RUN_P' => $first_run_p);

  ###
  ### Now what?
  ###

  ## If we actually collected all of the data and it passes muster,
  ## kick out to the data assembler.
  if( ! defined($bad_title_message) &&
      ! defined($bad_description_message) &&
      ! defined($bad_url_message) ){
    my $output_url = $self->url_for("/resource");
    $self->redirect_to($output_url->query('title' => $title,
					  'description' => $description,
					  'url' => $url));
    #die "kick";
  }else{
    ## Otherwise, keep looping back to the input page until we pass.
    $self->render('input');
  }
};

## Return QR code image for final wrenshoe location.
get '/qrcode.png' => sub {
  my $self = shift;

  my $title = $self->param('title') || undef;
  my $description = $self->param('description') || undef;
  my $url = $self->param('url') || undef;

  ## If these are not defined, kick back to /input.
  return if _insufficient_args_redirect($self, $title, $description, $url );

  my $qfinal_str = _bookmark($self, 'quiz', $title, $description, $url);

  ## TODO: Version needs to variable on the length of the string. May
  ## just want to enforce a max len.
  my $qr = GD::Barcode::QRcode->new($qfinal_str,
				    {Ecc=>'M', Version=>10, ModuleSize=>3});
  my $qrbytes = $qr->plot->png();

  #print STDERR $qrbytes . "\n";
  $self->render(data => $qrbytes, format => 'png');
};

## Check that build is possible with first pass assembly; give links,
## qrcode, stats, etc.
## Example URL:
##  http://localhost:3000/resource?title=Foo!&description=%3Cp%3E%3Cb%3EBar%3C%2Fb%3Eis+bar.%3C%2Fp%3E&url=https%3A%2F%2Fdocs.google.com%2Fspreadsheet%2Fccc%3Fkey%3D0AtdqMYPF4v3bdFZFRHJYbnlfX1o0Y1JZMmdhVy1vR3c%26usp%3Dsharing%23gid%3D0
get '/resource' => sub {
  my $self = shift;

  my $title = $self->param('title') || undef;
  my $description = $self->param('description') || undef;
  my $url = $self->param('url') || undef;

  ## If these are not defined, kick back to /input.
  return if _insufficient_args_redirect($self, $title, $description, $url );

  ## Attempt assembly from Google.
  my $ds = Wrenshoe::DeckSet->new();
  my $error_message = _assemble_attempt($ds, $url);
  if( defined($error_message) ){
    return if _error($self, 'assembly', 'Google Docs Error', $error_message);
  }

  ## Get the permanent resource URLs--the only things that change
  ## call-to-call besides this.
  my $qrcode_loc = _bookmark($self, 'qrcode.png', $title, $description, $url);
  my $quiz_loc = _bookmark($self, 'quiz', $title, $description, $url);

  $self->stash('DESCRIPTION' => $description);
  $self->stash('URL' => $url);
  $self->stash('QRCODE_URL' => $qrcode_loc);
  $self->stash('QUIZ_URL' => $quiz_loc);

  ## Render resource page.
  $self->render('resource');
};

## The fully dynamically generated inline wrenshoe quiz.
## Example URL:
##  http://localhost:3000/quiz?title=Foo!&description=%3Cp%3E%3Cb%3EBar%3C%2Fb%3Eis+bar.%3C%2Fp%3E&url=https%3A%2F%2Fdocs.google.com%2Fspreadsheet%2Fccc%3Fkey%3D0AtdqMYPF4v3bdFZFRHJYbnlfX1o0Y1JZMmdhVy1vR3c%26usp%3Dsharing%23gid%3D0
get '/quiz' => sub {
  my $self = shift;

  my $title = $self->param('title') || undef;
  my $description = $self->param('description') || undef;
  my $url = $self->param('url') || undef;

  ## If these are not defined, kick back to /input.
  return if _insufficient_args_redirect($self, $title, $description, $url );

  ## Attempt assembly from Google.
  my $ds = Wrenshoe::DeckSet->new();
  my $error_message = _assemble_attempt($ds, $url);
  if( defined($error_message) ){
    return if _error($self, 'assembly', 'Google Docs Error', $error_message);
  }

  ###
  ### Build the inline page.
  ###

  ##
  my $pgen = Wrenshoe::Page->new($ds);
  $pgen->title($title);
  $pgen->description($description);
  $pgen->inline_data_p(1);

  my $wrenshoe_style = 'bs3'; # TODO: make parameter
  my $tt_paths = [
		  'template/' . $wrenshoe_style,
		  'template/common',
		  '../template/' . $wrenshoe_style,
		  '../template/common'
		 ];

  ## A little OpenShift support if it smells right.
  my $oshd = $ENV{OPENSHIFT_HOMEDIR};
  if( $oshd ){
    push @$tt_paths,
      $oshd . '/app-root/runtime/repo/perl/template/' . $wrenshoe_style;
    push @$tt_paths,
      $oshd . '/app-root/runtime/repo/perl/template/common';
  }

  $pgen->template_paths($tt_paths);

  # my $here = getcwd();
  # my $abs_here = abs_path($here);
  # die "here: $abs_here";

  ## TODO: Render the final quiz.
  $self->render(text=>$pgen->page());
};

###
### Application start.
###

app->start();
